/*
	Original	:	Team Developer ParadisEO INRIA
	Complemento	:	Dagoberto Ramón Quevedo Orozco
	Fecha		:	28-Julio-2009
	Revisión	:	15-Abril-2010
	Institución	:	Universidad Autonoma de Nuevo Leon - CIDET - FIME
	Lenguaje	:	C++
*/

#include <eo>
#include <mo>
#include <fstream>
#include <iostream>
#include <string>
#include <time.h>
#include <ctime>
#include <stdlib.h>


clock_t startm, stopm;
#define START 		if((startm = clock()) == -1) {printf("Error calling clock");exit(1);}
#define STOP 		if((stopm  = clock()) == -1) {printf("Error calling clock");exit(1);}
#define PRINTTIME 	printf("%4.3f\n", ((double)stopm-startm)/CLOCKS_PER_SEC);

using namespace std;

static	int** 			A; // matriz A
static	int** 			B; // matriz B
static	int				n; // tamaño
static	int				k;
static	int 			z;

struct parametros
{
	string 		fileProblem;
	int 		TSmaxiItera;
	int 		TBsizeLista;
};

void 	liberarMemoria();
void	dateTimeSystem();
int		filenameNumber(string);
void 	Inicializacion(eoParser &, parametros &);

template <typename T>
string toString(T const&);

class QAProblema : public EO <eoMinimizingFitness>
{
	public:
		int*	solucion;

		QAProblema ()
	    {
	    	solucion = new int[n];
	    }
		
	  	QAProblema (const QAProblema & _problema)
    	{
    		solucion = new int[n];
      		for (int i = 0; i < n ; i++)
		  		solucion[i] = _problema.solucion[i];
	      	
			if (!_problema.invalid())
				fitness(_problema.fitness());
	    }

	  	~QAProblema()
	    {
	    	delete[] solucion;
	    }
		
	  	void operator= (const QAProblema & _problema)
	    {
	    	for ( int i = 0; i < n ; i++ )
		  		solucion[i] = _problema.solucion[i];

	      	fitness(_problema.fitness());
	    }

	  	void generar(int semilla)
	  	{
			cout<<"\nSemilla de aleatoriedad  : "<<semilla<<endl;
			srand(semilla);
	    	int r, t;
	   		for ( int i = 0; i < n; i++)
	      		solucion[i]=i;

	    	for ( int i = 0; i < n; i++ )
	      	{
				r= rand()%(n-i) + i;
				t= solucion[i];
				solucion[i] = solucion[r];
				solucion[r] = t;
	      	}
	  	}

	  	void imprimirSolucion()
	  	{
	    	for (int i = 0; i < n ; i++)
	      		cout<<solucion[i]+1<<(i!=n-1?",":"");
	  	}
};
 
class QAProblemaEval : public eoEvalFunc <QAProblema> 
{
	public:
	  	void operator() (QAProblema & _problema) {
			
			double z=0;
	    	for (int i = 0; i<n; i++ )
	      		for (int j = 0;j< n; j++ )
					z+=A[i][j]*B[_problema.solucion[i]][_problema.solucion[j]];

			_problema.fitness(z);
	  	}
};

class Movimiento : public moMove <QAProblema>
{
 public :
	int r;
	int s;

  	Movimiento()
	{ 
		r=s=0;
	}

	bool operator== (const Movimiento & _move)  const 
	{
		return (r == _move.r) && (s == _move.s); 
	}

  	void operator () (QAProblema & _problem)
  	{
    	int temp = _problem.solucion[r];
    	_problem.solucion[r] = _problem.solucion[s];
    	_problem.solucion[s] = temp;
  	}
};

class MovimientoInit : public moMoveInit <Movimiento> 
{
	public:
		void operator () (Movimiento & _mov, const QAProblema & _problema) {
			_mov.r = 0;
			_mov.s = 1;
	  	}
};

class MovimientoNext : public moNextMove <Movimiento> 
{
	public:
		bool operator() (Movimiento & _mov, const QAProblema & _problema) 
		{

			if (_mov.r<n-2)
			{
		  		if (_mov.s<n-1)
					_mov.s++;
				else 
				{
					_mov.r++;
					_mov.s=_mov.r+1;
				}
		  		return true;
			}
			return false;
		}
};

class MovimientoIncrEval : public moMoveIncrEval <Movimiento> 
{
	public:
		  int delta(int n, int** a, int** b, int* p, int r, int s)
		  {
			int w=0;
			for (k = 0; k < n; k = k + 1) 
				if (k!=r && k!=s)
					w+=(a[r][k]-a[s][k])*(b[p[s]][p[k]]-b[p[r]][p[k]]);
			
			return w*2;
		  }
		  
		  eoMinimizingFitness operator() (const Movimiento & _mov, const QAProblema & _p)
		  {
				return _p.fitness() + delta(n,A,B,_p.solucion,_mov.r, _mov.s);
		  }
};


int main(int argc, char **argv)
{
	if (argc<2){
		cout<<"Especifique un archivo de parametros" << endl;
		exit(1);
	}
	system("clear");
	
	parametros 				param;
	eoParser 				parser(argc,argv);	
	Inicializacion			(parser, param);
	QAProblema 				p;
  	QAProblemaEval 			problem_eval;
  	MovimientoNext 			move_next;
  	MovimientoIncrEval 		move_incr_eval;
	MovimientoInit 			move_init;
  	moBestImprSelect		<Movimiento> move_select;
	moSimpleMoveTabuList	<Movimiento> tabulist(param.TBsizeLista);
  	moImprBestFitAspirCrit	<Movimiento> aspiration_criterion;
  	moGenSolContinue		<QAProblema> continu (param.TSmaxiItera);
  	moTS 					<Movimiento> TS (move_init, move_next, move_incr_eval, 
							 				tabulist, aspiration_criterion, continu, problem_eval);

	cout<<endl;
	cout<<"\nArchivo de problema      : "<<param.fileProblem.c_str();	
	cout<<"\nTamaño de problema       : "<<n;
	cout<<"\nMaximo # de iteraciones  : "<<param.TSmaxiItera;
	cout<<"\nTamaño de Lista Tabu     : "<<param.TBsizeLista;
	cout<<"\nMejor resultado conocido : "<<k;
	START;
	p.generar(filenameNumber(param.fileProblem.c_str()));
	problem_eval(p);
	cout<<"\nFecha-Hora de inicio     : ";dateTimeSystem();
	cout<<"\nSolución inicial         : ";p.imprimirSolucion();
	printf("=%10.0Lf\n", (long double)p.fitness());
	TS(p);
	STOP;
	cout<<"\nFecha-Hora final         : ";dateTimeSystem();
	cout<<"\nSolución final           : ";p.imprimirSolucion();
	printf("=%10.0Lf\n", (long double)p.fitness());
	cout<<"\nTiempo de ejecución (seg): ";
	PRINTTIME;
	liberarMemoria();
	return 0;
}

void Inicializacion(eoParser & parser, parametros & param)
{
	param.fileProblem 	= parser.createParam(string("")		, "fileProblem"	, "",'i', "" ).value();
	param.TSmaxiItera 	= parser.createParam(unsigned(100)	, "TSmaxiItera"	, "",'M', "" ).value();
	param.TBsizeLista 	= parser.createParam(unsigned(100)	, "TBsizeLista" , "",'l', "" ).value();

	ifstream	archivo_data;
	ifstream	archivo_soln;
	int 		i, j;
	archivo_data.open(param.fileProblem.c_str());
	archivo_soln.open((param.fileProblem.substr(0,param.fileProblem.size()-4)+".sln").c_str());

	if (!archivo_data.is_open() || !archivo_soln.is_open() )
	{
		cout<<"\nError al intentar leer archivo de datos "<<param.fileProblem.c_str()<<". Verifique su existencia\n";
		exit(1);
	}

	archivo_soln>>k;
	archivo_soln>>k;
	archivo_data>>n;

	A = new int* [n];
	B = new int* [n];

	for (i=0; i<n; i++)
	{
		A[i] = new int[n];
		B[i] = new int[n];
	}


	for (i=0; i<n; i++)
		for (j=0; j<n; j++) archivo_data>>A[i][j];

	for (i=0; i<n; i++)
		for (j=0; j<n; j++) archivo_data>>B[i][j];

	archivo_data.close();
}

void dateTimeSystem()
{
	struct tm	*current;
	time_t 		now;
	time(&now);
	current = 	localtime(&now);

	cout<<current->tm_year+1900<<"/"<<(current->tm_mon <10?"0":"")<<current->tm_mon
							   <<"/"<<(current->tm_mday<10?"0":"")<<current->tm_mday
                               <<" "<<(current->tm_hour<10?"0":"")<<current->tm_hour
   							   <<":"<<(current->tm_min <10?"0":"")<<current->tm_min
							   <<":"<<(current->tm_sec <10?"0":"")<<current->tm_sec;
}

//Funciones Diversas

int filenameNumber(string file)
{
	int sum=0;
	for(int w=0;w<file.size();w++)
		sum+=(int)file[w];
	return sum;
}

void liberarMemoria()
{
	for (int i = 0; i < n; i++)
	{
		delete[] A[i];
		delete[] B[i];
	}

	delete[] A;
	delete[] B;
}

template <typename T>
string toString(T const& value) {
    stringstream s;
    s<<value;
    return s.str();
}
